# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tchannel import TChannel

from yarpc.transport.http import HTTPOutbound
from yarpc.transport.tchannel import TChannelOutbound


HTTP_PORT = 8081
TCHANNEL_PORT = 8082


def factory(server, transport_name):
    if transport_name == 'http':
        return HTTPOutbound('http://%s:%d' % (server, HTTP_PORT))
    elif transport_name == 'tchannel':
        ch = TChannel('client')
        hostport = '%s:%d' % (server, TCHANNEL_PORT)
        return TChannelOutbound(ch, hostport)
    raise Exception('Unrecognized transport_name: %s' % transport_name)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen
from tchannel import TChannel, Response
from tchannel.errors import (
    BadRequestError as TChBadRequestError,
    UnexpectedError as TChUnexpectedError,
)
from tchannel.sync import TChannel as TChannelSync

from yarpc.errors import (
    BadRequestError,
    UnexpectedError,
    ServerEncodingError,
    ProcedureFailedError,
)
from yarpc.transport import Inbound, Request
from yarpc.transport.validator import _Validator
from yarpc._future import fail_to, maybe_async
from . import headers


class TChannelInbound(Inbound):

    def __init__(self, tchannel):
        if isinstance(tchannel, TChannelSync):
            raise ValueError(
                "TChannelInbound must use a regular TChannel,"
                "do not use a TChannelSync"
            )
        self._tchannel = tchannel

    def start(self, handler):
        self._tchannel.register(TChannel.FALLBACK)(_TChannelHandler(handler))
        if not self._tchannel.is_listening():
            self._tchannel.listen()

    def stop(self):
        self._tchannel.close()

    @property
    def hostport(self):
        return self._tchannel.hostport

    @property
    def port(self):
        return int(self.hostport.rsplit(':', 1)[1])


class _TChannelHandler(object):

    def __init__(self, handler):
        self.handler = handler

    def __call__(self, tchan_request):
        request = _to_request(tchan_request)

        v = _Validator(request)
        try:
            v.validate()
        except BadRequestError as e:
            raise TChBadRequestError(
                'BadRequest: %s\n' % str(e)
            )

        answer = gen.Future()

        @fail_to(answer)
        def on_dispatch(future):
            if not future.exception():
                response = future.result()
                tchan_response = Response(
                    body=response.body,
                    headers=headers.encode(
                        request.encoding, response.headers),
                    transport=None,  # TODO transport?
                )
                answer.set_result(tchan_response)
                return

            _handle_request_exception(future.exception(), request)

        maybe_async(
            self.handler.handle, request
        ).add_done_callback(on_dispatch)

        return answer


def _handle_request_exception(e, request):
    if isinstance(e, BadRequestError):
        raise TChBadRequestError(
            'BadRequest: %s\n' % str(e)
        )

    if isinstance(e, ServerEncodingError):
        raise e

    if not isinstance(e, UnexpectedError):
        e = ProcedureFailedError(
            exception=e,
            service=request.service,
            procedure=request.procedure,
        )

    m = 'UnexpectedError: %s\n' % str(e)

    raise TChUnexpectedError(m)


def _to_request(tchan_request):
    scheme = tchan_request.transport.scheme

    request = Request(
        caller=tchan_request.transport.caller_name,
        service=tchan_request.service,
        encoding=scheme,
        ttl=tchan_request.timeout,  # in seconds
        procedure=tchan_request.endpoint,
        headers=headers.decode(scheme, tchan_request.headers),
        body=tchan_request.body,
    )

    return request



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module provides the Thrift encoding for YARPC.
"""

# TODO usage examples in docstring

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from types import ModuleType

import thriftrw

from yarpc.errors import ValueExpectedError, OneWayNotSupportedError
from yarpc.transport import Response
from yarpc._future import reraise
from ._encoding import OutboundEncoder, InboundEncoder
from ._async import AsyncClient, AsyncHandler


NAME = 'thrift'

# TODO stuttering? do we really want everything to have Thrift in front of its
# name even though we have namespaces?

# TODO perhaps we should introduce a Thrift-only Request type which makes it
# impossible to set the procedure.

# TODO Move the load(...) logic and types into a non-public submodule of this
# package.


def load(path):
    """Loads the Thrift file at the given path."""
    return _wrap(thriftrw.load(path))


def _wrap(thriftrw_module):
    """Wraps a thriftrw module to generate requests for YARPC."""

    module = ModuleType(thriftrw_module.__name__)

    # Attach services
    for s in thriftrw_module.__services__:
        setattr(module, s.service_spec.name, Service(module, s))

    # Attach types
    for t in thriftrw_module.__types__:
        # Workaround for thriftrw/thriftrw-python#124
        if hasattr(t, '__name__'):
            setattr(module, t.__name__, t)

    # Attach constants
    for n, c in thriftrw_module.__constants__.items():
        setattr(module, n, c)

    # Attach includes
    for m in thriftrw_module.__includes__:
        setattr(module, m.__name__, m)

    return module


class Service(object):

    def __init__(self, module, service_cls):
        self.service_cls = service_cls
        self.service_spec = service_cls.service_spec

        for func_spec in self.service_spec.functions:
            setattr(self, func_spec.name, Function(module, self, func_spec))

    @property
    def name(self):
        return self.service_spec.name

    def __str__(self):
        return 'Service(%s)' % self.name

    __repr__ = __str__


class Function(object):

    def __init__(self, module, service, func_spec):
        self.spec = func_spec
        self.service = service

        function = func_spec.surface
        self._request_cls = function.request
        self._response_cls = function.response

    @property
    def name(self):
        return '%s::%s' % (self.service.name, self.spec.name)

    def __str__(self):
        return 'Function(%s)' % self.name

    __repr__ = __str__

    def __call__(self, *args, **kwargs):
        if self.spec.oneway:
            raise OneWayNotSupportedError(
                'YARPC does not currently support oneway procedures.'
            )
        return self._request_cls(*args, **kwargs)


class ThriftClient(object):
    """ThriftClient builds a client that sends Thrift requests via the given
    channel.

    .. code-block:: python

        foo = thrift.load('foo.thrift')
        client = thrift.ThriftClient(rpc.channel('foo'))

        response = yield client.call(Request(
            headers={'foo': 'bar'},
            body=foo.MyService.someMethod(arg1, arg2),
        ))
    """

    __slots__ = ('_impl',)

    def __init__(self, channel):
        self._impl = AsyncClient(channel)

    @property
    def channel(self):
        return self._impl.channel

    def call(self, request):
        encoder = _OutboundEncoder()
        return self._impl.call(encoder, request)


class _OutboundEncoder(OutboundEncoder):

    name = 'thrift'

    def __init__(self):
        self.module = None
        self.result_spec = None

    def encode_request(self, request):
        thrift_request = request.body
        request_cls = thrift_request.__class__
        function_spec = request_cls.type_spec.function
        service_spec = function_spec.service

        self.module = thrift_request.__thrift_module__
        self.result_spec = request_cls.result_type.type_spec

        request.encoding = NAME
        request.procedure = _get_procedure_name(
            service=service_spec.name,
            function=function_spec.name,
        )

        request.body = self.module.dumps(request.body)
        return request

    def decode_response(self, response):
        thrift_response = self.module.loads(
            self.result_spec.surface,
            response.body,
        )
        response.body = _unwrap_body(
            response_spec=self.result_spec,
            body=thrift_response,
        )
        return response


class _InboundEncoder(InboundEncoder):

    name = 'thrift'

    def __init__(self, function_spec):
        self.function_spec = function_spec
        self.args_spec = function_spec.args_spec
        self.result_spec = function_spec.result_spec
        self.module = self.args_spec.surface.__thrift_module__

    def decode_request(self, request):
        request.body = self.module.loads(self.args_spec.surface, request.body)
        return request

    def encode_response(self, request, response):
        if response.body is None:
            # TODO separate exception
            assert self.result_spec.return_spec is None, (
                'Expected a value to be returned for %s, '
                'but recieved None - only void procedures can '
                'return None.' % self.function_spec.name
            )
            response.body = self.result_spec.surface()
        else:
            response.body = self.result_spec.surface(success=response.body)
        response.body = self.module.dumps(response.body)
        return response

    def encode_exc_info(self, exc_info):
        response_body = None

        e = exc_info[1]
        for exc_spec in self.result_spec.exception_specs:
            if isinstance(e, exc_spec.spec.surface):
                response_body = self.result_spec.surface(**{exc_spec.name: e})
                break

        if response_body is not None:
            response_body = self.module.dumps(response_body)
            return Response(body=response_body)

        # No match. Re-raise.
        reraise(exc_info)


class ThriftHandler(object):

    __slots__ = ('_impl', 'function_spec')

    def __init__(self, function_spec, handler):
        self.function_spec = function_spec
        self._impl = AsyncHandler(handler)

    @property
    def handler(self):
        return self._impl.handler

    def handle(self, request):
        encoder = _InboundEncoder(self.function_spec)
        return self._impl.handle(encoder, request)


class ThriftProcedure(object):

    def __init__(self, service_spec, function_spec, handler):
        self.name = _get_procedure_name(
            service=service_spec.name,
            function=function_spec.name,
        )
        self.handler = handler

        self.transport_handler = ThriftHandler(function_spec, handler)

    @property
    def procedures(self):
        return {self.name: self.transport_handler}

    def __call__(self, *args, **kwargs):
        return self.handler(*args, **kwargs)


def procedure(service, method=None):
    """Builds a Thrift procedure.

    .. code-block:: python

        from yarpc.encoding import thrift

        foo = thrift.load('foo.thrift')

        @thrift.procedure(foo.MyService)
        def someMethod(request):
            # ...
    """

    def decorator(f):
        function = method
        if function is None:
            function = f.__name__
        function = getattr(service, function)
        # TODO raise more user friendly exception on AttributeError.
        return ThriftProcedure(service.service_spec, function.spec, f)

    return decorator


def _get_procedure_name(service, function):
    return '%s::%s' % (service, function)


def _unwrap_body(response_spec, body):

    # exception - reraise
    for exc_spec in response_spec.exception_specs:
        exc = getattr(body, exc_spec.name)
        if exc is not None:
            raise exc

    # success - non-void
    if response_spec.return_spec is not None:
        if body.success is None:
            raise ValueExpectedError(
                'Expected a value to be returned in %s, but received '
                'None. Only void procedures can return None.'
                % str(response_spec.surface)
            )

        return body.success

    # success - void
    else:
        return None



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module provides the JSON encoding for YARPC.
"""

# TODO usage examples in docstring

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import json

from ._encoding import SimpleEncoder
from ._async import AsyncClient, AsyncHandler


# TODO stuttering? do we really want everything to have JSON in front of its
# name even though we have namespaces?
# TODO limit the surface area of the exported API

_encoder = json.JSONEncoder(separators=(',', ':'))


def _encode(s):
    s = _encoder.encode(s)
    s += '\n'
    return s

#: JSONEncoder uses JSON to encode/decode request bodies.
JSONEncoder = SimpleEncoder(
    name='json',
    # this lambda removes the spacing in json objects
    encode=_encode,
    decode=json.loads,
)


class JSONClient(object):
    """JSONClient builds a JSON client that sends requests via the given
    channel.

    .. code-block:: python

        client = JSONClient(rpc.channel('myservice'))
        response = yield client.call(Request(
            procedure='hello',
            headers={'token': 'foo'},
            body={'name': 'somecaller'},
        ))
    """

    __slots__ = ('_impl',)

    def __init__(self, channel):
        self._impl = AsyncClient(channel)

    @property
    def channel(self):
        return self._impl.channel

    def call(self, request):
        return self._impl.call(JSONEncoder, request)


class JSONHandler(object):
    """JSONHandler wraps a JSON procedure implementation into a transport
    level asynchronous Handler.
    """

    __slots__ = ('_impl',)

    def __init__(self, handler):
        self._impl = AsyncHandler(handler)

    @property
    def handler(self):
        return self._impl.handler

    def handle(self, request):
        return self._impl.handle(JSONEncoder, request)


class JSONProcedure(object):

    def __init__(self, name, handler):
        self.name = name

        # Handler is what the user gave us. We need to retain it so that they
        # can call it directly. transport_handler is what will get registered.
        self.handler = handler
        self.transport_handler = JSONHandler(handler)

    @property
    def procedures(self):
        return {self.name: self.transport_handler}

    def __call__(self, *args, **kwargs):
        return self.handler(*args, **kwargs)


def procedure(name):
    """Builds a JSON procedure. TODO explain.

    .. code-block:: python

        from yarpc.encoding import json

        @json.procedure('hello')
        def hello(request):
            print(request.headers)
            return Response(body={
                'success': 'hello, %s' % request.body['name']
            })

        rpc.register(hello)

    Note that the decorated function can still be called directly with a JSON
    request.

    .. code-block:: python

        hello(Request(body={'foo': 'bar'}))
    """

    def decorator(f):
        return JSONProcedure(name, f)

    return decorator



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import logging
import os
import sys

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


def async_main(start_ioloop=False):
    import tornado.ioloop
    from async import client, server

    server.start()
    client.start()
    if start_ioloop:
        tornado.ioloop.IOLoop.current().start()


def sync_main():
    from sync import client

    client.start()


if __name__ == '__main__':
    if os.environ.get('SYNC', False) is False:
        async_main(start_ioloop=True)
    else:
        sync_main()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import json
import struct
from io import BytesIO

from tchannel.schemes import JSON


def encode(scheme, headers):
    headers = headers or {}
    if scheme == JSON:
        for k, v in headers.iteritems():
            headers[k] = str(v)
        return json.dumps(headers)
    return _encode(headers)


def decode(scheme, src):
    if not src:
        return {}
    if scheme == JSON:
        return _normalize(json.loads(src))
    return _normalize(_decode(BytesIO(src)))


def _normalize(headers):
    if headers:
        return {k.lower(): v for k, v in headers.iteritems()}
    return {}


def _encode(headers):
    """Writes headers in the format ``nh:2 (k~2 v~2){nh}``."""

    out = BytesIO()
    _put_i16(out, len(headers))  # nh:2

    for k, v in headers.items():
        _put_str16(out, k)  # k~2
        _put_str16(out, v)  # v~2

    return out.getvalue()


def _decode(src):
    """Readers headers in the format ``nh:2 (k~2 v~2){nh}``."""

    num = _get_i16(src)
    headers = {}

    for i in xrange(num):
        k = _get_str16(src)
        v = _get_str16(src)
        headers[k] = v

    return headers


def _put_str16(out, s):
    """Writes a string in the format ``len(s):2 s:*``"""
    s = str(s)
    _put_i16(out, len(s))  # len(s):2
    out.write(s)


def _get_str16(src):
    """Reads a string in the format ``len(s):2 s:*``"""
    length = _get_i16(src)
    s = src.read(length)
    assert len(s) == length, (
        'expected %d bytes, got %d' % (length, len(s))
    )  # TODO explicit decode exception
    return s


def _put_i16(out, i):
    """Writes a 16-bit integer in big-endian byte-order."""
    out.write(struct.pack('>H', i))


def _get_i16(src):
    """Reades a 16-bit integer in big-endian byte-order."""
    s = src.read(2)
    assert len(s) == 2, 'ran out of input'  # TODO explicit decode exception
    return struct.unpack('>H', s)[0]



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen
import tornado.web
import tornado.httpserver
import tornado.netutil

from yarpc.errors import BadRequestError, UnexpectedError, ProcedureFailedError
from yarpc.transport import Inbound, Request
from yarpc.transport.validator import _Validator
from yarpc._future import fail_to, maybe_async
from . import headers


class HTTPInbound(Inbound):

    def __init__(self, port=None, reuse_port=False, handlers=None):
        self.port = port or 0
        self._reuse_port = reuse_port
        self._server = None
        self._handlers = handlers or []

    def start(self, handler):
        self._handlers.append(
            ('.*$', _TornadoHandler, dict(handler=handler)),
        )
        app = tornado.web.Application(handlers=self._handlers)
        self._server = tornado.httpserver.HTTPServer(app)

        # the following does the same as http_server.listen(self.port),
        # except with fine grain control over the sockopts
        sockets = tornado.netutil.bind_sockets(
            port=self.port,
            reuse_port=self._reuse_port,
        )
        self.port = sockets[0].getsockname()[1]
        self._server.add_sockets(sockets)

    def stop(self):
        if self._server is not None:
            self._server.stop()
            self._server = None

    @property
    def hostport(self):
        # TODO get pub ip???
        return 'localhost:%s' % self.port


class _TornadoHandler(tornado.web.RequestHandler):

    def initialize(self, handler):
        self._handler = handler

    def post(self):
        request = _to_request(self.request)

        if not self._validate(request):
            return

        answer = gen.Future()

        @fail_to(answer)
        def on_dispatch(future):
            if not future.exception():
                response = future.result()
                if response.headers:
                    hs = headers.to_http_headers(response.headers)
                    for k, v in hs.iteritems():
                        self.set_header(k, v)
                if response.body is not None:
                    self.write(response.body)
            else:
                self._handle_request_exception(future.exception(), request)

            answer.set_result(None)

        maybe_async(
            self._handler.handle, request
        ).add_done_callback(on_dispatch)

        return answer

    def _validate(self, request):
        v = _Validator(request)
        if request.ttl:
            v.parse_ttl(request.ttl)
        try:
            v.validate()
        except BadRequestError as e:
            m = 'BadRequest: %s\n' % str(e)
            self.write(m)
            self.set_status(400)
            return False
        return True

    def _handle_request_exception(self, e, request):
        if isinstance(e, BadRequestError):
            m = 'BadRequest: %s\n' % str(e)
            self.write(m)
            self.set_status(400)
            return

        if not isinstance(e, UnexpectedError):
            e = ProcedureFailedError(
                exception=e,
                service=request.service,
                procedure=request.procedure,
            )

        m = 'UnexpectedError: %s\n' % str(e)
        self.write(m)
        self.set_status(500)


def _to_request(http_request):
    http_headers = http_request.headers
    procedure = http_headers.pop(headers.PROCEDURE, None)
    caller = http_headers.pop(headers.CALLER, None)
    service = http_headers.pop(headers.SERVICE, None)
    ttl = http_headers.pop(headers.TTL, None)
    encoding = http_headers.pop(headers.ENCODING, None)  # optional

    request = Request(
        caller=caller,
        service=service,
        encoding=encoding,
        ttl=ttl,
        procedure=procedure,
        headers=headers.from_http_headers(http_headers),
        body=http_request.body,
    )

    return request



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from yarpc.errors import MissingParametersError, InvalidTTLError


class _Validator(object):
    def __init__(self, request):
        self._request = request
        self._error = None

    def parse_ttl(self, ttl):
        if ttl == '':
            self._request.ttl = 0
            return
        try:
            self._request.ttl = int(ttl)
        except (TypeError, ValueError):
            self._error = InvalidTTLError(
                service=self._request.service,
                procedure=self._request.procedure,
                ttl=ttl,
            )

    def validate(self):
        # already failed
        if self._error:
            raise self._error

        # check missing params
        missing = []
        if not self._request.service:
            missing.append('service name')
        if not self._request.procedure:
            missing.append('procedure')
        if not self._request.caller:
            missing.append('caller name')
        if not self._request.ttl:
            missing.append('TTL')
        if not self._request.encoding:
            missing.append('encoding')
        if missing:
            raise MissingParametersError(missing)

        # negative TTLs are invalid
        if self._request.ttl < 0:
            raise InvalidTTLError(
                service=self._request.service,
                procedure=self._request.procedure,
                ttl=self._request.ttl,
            )

        return self._request



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module provides the Raw encoding for YARPC.
"""

# TODO usage examples in docstring

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from ._encoding import SimpleEncoder
from ._async import AsyncClient, AsyncHandler


# TODO stuttering in naming.
# TODO limit the surface area of the exported API

def _identity(x):
    return x

#: RawEncoder provides the Raw encoding format.
RawEncoder = SimpleEncoder('raw', _identity, _identity)


class RawClient(object):
    """RawClient builds a client that sends raw requests via the given
    channel.

    .. code-block:: python

        client = RawClient(rpc.channel('myservice'))
        response = yield client.call(Request(
            procedure='hello',
            body='something',
        ))
    """

    __slots__ = ('_impl',)

    def __init__(self, channel):
        self._impl = AsyncClient(channel)

    @property
    def channel(self):
        return self._impl.channel

    def call(self, request):
        return self._impl.call(RawEncoder, request)


class RawHandler(object):
    """RawHandler wraps a raw procedure implementation into a transport
    level asynchronous Handler.
    """

    __slots__ = ('_impl',)

    def __init__(self, handler):
        self._impl = AsyncHandler(handler)

    @property
    def handler(self):
        return self._impl.handler

    def handle(self, request):
        return self._impl.handle(RawEncoder, request)


class RawProcedure(object):

    def __init__(self, name, handler):
        self.name = name

        # Handler is what the user gave us. We need to retain it so that they
        # can call it directly. transport_handler is what will get registered.
        self.handler = handler
        self.transport_handler = RawHandler(handler)

    @property
    def procedures(self):
        return {self.name: self.transport_handler}

    def __call__(self, *args, **kwargs):
        return self.handler(*args, **kwargs)


def procedure(name):
    """Builds a raw procedure. TODO explain.

    .. code-block:: python

        from yarpc.encoding import raw

        @raw.procedure('hello')
        def hello(request):
            return Response(body='foo')

        rpc.register(hello)

    Note that the decorated function can still be called directly with a raw
    request.

    .. code-block:: python

        hello(Request(body='foo'))
    """

    def decorator(f):
        return RawProcedure(name, f)

    return decorator



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from yarpc.sync import RPC

from crossdock import transport as tr
from crossdock.async.client.behavior.headers import (
    tests, get_caller
)


def headers(respw, server, transport, encoding, **kwargs):
    rpc = RPC(
        service='client',
        outbounds={
            'yarpc-test': tr.factory(server, transport),
        },
    )
    caller = get_caller(encoding, rpc.channel('yarpc-test'))

    for test in tests:
        desc, give, want = test
        try:
            resp = caller.call(give).result()
        except Exception as e:
            respw.fail('%s: call failed, %s' % (desc, str(e)))
        else:
            got = resp.headers
            if want != got:
                respw.fail(
                    '%s: call failed, got %s, want %s' % (desc, got, want)
                )
            else:
                respw.success('%s: returns valid headers' % desc)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module provides common, encoding-agnostic functionality for Tornado. It
should not be used outside the encoding package.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen

from yarpc.errors import (
    ServerDecodingError,
    ServerEncodingError,
    ProcedureFailedError,
    EncodingMismatchError,
)
from yarpc._future import transform, fail_to, maybe_async
from yarpc import transport


class AsyncClient(object):
    """Encoding agnostic asynchronous client.

    :param channel:
        Channel through which requests will be sent.
    """

    def __init__(self, channel):
        self.channel = channel

    def call(self, encoder, request):
        request = encoder.encode_request(request)
        future = self.channel.call(request)
        return transform(future, encoder.decode_response)


class AsyncHandler(transport.Handler):
    """Encoding agnsotic asynchronous handler.

    Wraps a handler of the given encoding type to speak raw bytes and act as a
    transport-level handler.

    :param handler:
        A function that accepts a request and returns its response or a future
        that resolves to the response.
    """

    def __init__(self, handler):
        self.handler = handler
        # NOTE: Transport-level Handlers have a handle() method but
        # encoding-level handlers are just functions that accept a Request and
        # return a Future<Response>.

    def handle(self, encoder, request):
        if request.encoding != encoder.name:
            raise ServerDecodingError(
                exception=EncodingMismatchError(
                    want=encoder.name,
                    got=request.encoding,
                ),
                encoding=encoder.name,
                service=request.service,
                procedure=request.procedure,
                caller=request.caller,
            )

        try:
            request = encoder.decode_request(request)
        except Exception as e:
            raise ServerDecodingError(
                exception=e,
                encoding=encoder.name,
                service=request.service,
                procedure=request.procedure,
                caller=request.caller,
            )

        answer = gen.Future()

        @fail_to(answer)
        def on_done(future):
            try:
                if future.exception():
                    response = encoder.encode_exc_info(future.exc_info())
                else:
                    response = encoder.encode_response(
                        request,
                        future.result(),
                    )
            except ServerEncodingError:
                raise
            except Exception as e:
                raise ProcedureFailedError(
                    service=request.service,
                    procedure=request.procedure,
                    exception=e,
                )
            answer.set_result(response)

        maybe_async(self.handler, request).add_done_callback(on_done)
        return answer



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module provides primitives for YARPC.

Types defined in this file MUST be agnostic of whether we're synchronous or
asynchornous.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from .validator import _Validator


class Request(object):

    def __init__(
        self,
        caller=None,
        service=None,
        encoding=None,
        ttl=None,
        procedure=None,
        headers=None,
        body=None,
    ):
        self.caller = caller
        self.service = service
        self.encoding = encoding

        # TODO switch to timedelta
        # TODO use same default as spec/other langs
        self.ttl = ttl

        self.procedure = procedure
        self.headers = headers
        self.body = body

        # TODO maybe we want to create separate application-level Request
        # objects so that users don't accidentally set caller/service/encoding
        # since those are filled at a higher level


class Response(object):

    def __init__(self, headers=None, body=None):
        self.headers = headers
        self.body = body


class Outbound(object):

    __slots__ = ()

    def call(self, request):
        """Sends the given request over the wire and returns its response."""
        raise NotImplementedError


class Inbound(object):

    __slots__ = ()

    def start(self, handler):
        """Serves the given Handler over this Inbound.

        This MUST return immediately, although it SHOULD block until the
        Inbound is ready to accept requests.
        """
        raise NotImplementedError

    def stop(self):
        """Stops serving this Inbound.

        This MAY block while the server drains ongoing requests.
        """
        raise NotImplementedError


class Handler(object):

    __slots__ = ()

    def handle(self, request):
        """Handles the given request.

        Returns the Response for the given Request.
        """
        raise NotImplementedError


class Channel(object):
    """A Channel scopes an Outbound to a single caller-service pair.

    Channel-local settings take precedence over global RPC-level settings.
    """

    __slots__ = ('caller', 'service', 'outbound')

    def __init__(self, caller, service, outbound):
        self.caller = caller
        self.service = service
        self.outbound = outbound

        # TODO channel-local timeouts, etc.

    def call(self, request):
        # TODO do we always want to stomp the caller and service?
        request.caller = request.caller or self.caller
        request.service = request.service or self.service

        v = _Validator(request)
        v.validate()

        # TODO apply channel-local timeouts

        return self.outbound.call(request)


class Registrant(object):
    """Registrant provides zero or more procedures and their handlers."""

    __slots__ = ()

    @property
    def procedures(self):
        """Map from procedure name to Handler for that procedure."""
        raise NotImplementedError


class Procedure(Registrant, Handler):
    """Procedure is a Registrant that provides a single procedure and adapts a
    function which accepts a Request and returns a Response into a Handler.
    """

    __slots__ = ('name', 'handler')

    def __init__(self, name, handler):
        self.name = name
        self.handler = handler

    @property
    def procedures(self):
        return {self.name: self}

    def handle(self, request):
        return self.handler(request)

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self.name)

    __repr__ = __str__



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from os import path

from tornado import gen

import crossdock
from crossdock import rand, transport as tr
from yarpc import RPC, Request
from yarpc.encoding.raw import RawClient
from yarpc.encoding.json import JSONClient
from yarpc.encoding.thrift import load, ThriftClient

idl = path.join(path.dirname(crossdock.__file__), 'Echo.thrift')
service = load(idl)

token1 = rand.string(10)
token2 = rand.string(10)

# tuples of (desc, give, want)
tests = [
    (
        'valid headers',
        {'token1': token1, 'token2': token2},
        {'token1': token1, 'token2': token2},
    ),
    (
        'non-string values',
        {'token': 42},
        {'token': '42'},
    ),
    (
        'empty strings',
        {'token': ''},
        {'token': ''},
    ),
    (
        'no headers',
        None,
        {},
    ),
    (
        'empty map',
        {},
        {},
    ),
    (
        'varying casing',
        {'ToKeN1': token1, 'tOkEn2': token2},
        {'token1': token1, 'token2': token2},
    ),
    (
        'http header conflict',
        {'Rpc-Procedure': 'does not exist'},
        {'rpc-procedure': 'does not exist'},
    ),
    (
        'mixed case value',
        {'token': 'MIXED case Value'},
        {'token': 'MIXED case Value'},
    ),
]


@gen.coroutine
def headers(respw, server, transport, encoding, **kwargs):
    rpc = RPC(
        service='client',
        outbounds={
            'yarpc-test': tr.factory(server, transport),
        },
    )
    caller = get_caller(encoding, rpc.channel('yarpc-test'))

    for test in tests:
        desc, give, want = test
        try:
            resp = yield caller.call(give)
        except Exception as e:
            respw.fail('%s: call failed, %s' % (desc, str(e)))
        else:
            got = resp.headers
            if want != got:
                respw.fail(
                    '%s: call failed, got %s, want %s' % (desc, got, want)
                )
            else:
                respw.success('%s: returns valid headers' % desc)


def get_caller(encoding, channel):
    if encoding == 'raw':
        return RawCaller(channel)
    elif encoding == 'json':
        return JSONCaller(channel)
    elif encoding == 'thrift':
        return ThriftCaller(channel)
    else:
        raise ValueError('encoding must be either raw, json, or thrift')


class RawCaller(object):

    def __init__(self, channel):
        self._client = RawClient(channel)

    def call(self, headers):
        request = Request(
            procedure='echo/raw',
            ttl=10000,
            headers=headers,
        )
        return self._client.call(request)


class JSONCaller(object):

    def __init__(self, channel):
        self._client = JSONClient(channel)

    def call(self, headers):
        request = Request(
            procedure='echo',
            ttl=10000,
            headers=headers,
        )
        return self._client.call(request)


class ThriftCaller(object):

    def __init__(self, channel):
        self._client = ThriftClient(channel)

    def call(self, headers):
        request = Request(
            body=service.Echo.echo(service.Ping(beep='boop')),
            ttl=10000,
            headers=headers,
        )
        return self._client.call(request)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)


class YARPCError(Exception):
    pass


class OneWayNotSupportedError(YARPCError):
    pass


class ValueExpectedError(YARPCError):
    pass


class NoOutboundForServiceError(YARPCError):
    pass


class UnexpectedError(YARPCError):
    pass


class BadRequestError(YARPCError):
    pass


class ProcedureFailedError(UnexpectedError):

    def __init__(self, exception, service, procedure):
        message = (
            'error for procedure "%s" of service "%s": %s' %
            (procedure, service, str(exception))
        )
        super(ProcedureFailedError, self).__init__(message)


class UnrecognizedProcedureError(BadRequestError):

    def __init__(self, procedure, service):
        message = (
            'unrecognized procedure "%s" for service "%s"'
            % (procedure, service)
        )
        super(UnrecognizedProcedureError, self).__init__(message)


class MissingParametersError(BadRequestError):

    def __init__(self, params):
        message = self._construct_message(params)
        super(MissingParametersError, self).__init__(message)

    @staticmethod
    def _construct_message(params):
        message = 'missing '
        if len(params) == 1:
            message += params[0]
            return message

        if len(params) == 2:
            message += '%s and %s' % (params[0], params[1])
            return message

        message += ', '.join(params[:-1])
        message += ', and %s' % params[-1]
        return message


class InvalidTTLError(BadRequestError):

    def __init__(self, service, procedure, ttl):
        message = (
            'invalid TTL "%s" for procedure "%s" of service "%s": '
            'must be positive integer' % (ttl, procedure, service)
        )
        super(InvalidTTLError, self).__init__(message)


class ServerDecodingError(BadRequestError):

    def __init__(self, exception, encoding, service, procedure, caller):
        message = (
            'failed to decode "%s" request body for procedure "%s" of '
            'service "%s" from caller "%s": %s' %
            (encoding, procedure, service, caller, str(exception))
        )
        super(ServerDecodingError, self).__init__(message)


class ServerEncodingError(UnexpectedError):

    def __init__(self, exception, encoding, service, procedure, caller):
        message = (
            'failed to encode "%s" response body for procedure "%s" '
            'of service "%s" from caller "%s": %s' %
            (encoding, procedure, service, caller, str(exception))
        )
        super(ServerEncodingError, self).__init__(message)


class EncodingMismatchError(BadRequestError):

    def __init__(self, want, got):
        message = 'expected encoding "%s" but got "%s"' % (want, got)
        super(EncodingMismatchError, self).__init__(message)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError

from yarpc.errors import BadRequestError, UnexpectedError
from yarpc.transport import Outbound, Response
from yarpc._future import fail_to, reraise
from . import headers


class HTTPOutbound(Outbound):
    """

    :param str url:
        URL to which HTTP requests will be sent.
    :param AsyncHTTPClient client:
        HTTP client used to make requests. A default client will be
        constructed if omitted.
    """

    def __init__(self, url, client=None):
        self.url = url

        # We don't build the default client here because if we're using
        # threadloop, we want the IOLoop of the threadloop to be used by the
        # AsyncHTTPClient but it resolves the IOLoop at instantiation.
        self.client = client

    def call(self, request):
        if self.client is None:
            self.client = AsyncHTTPClient()

        answer = gen.Future()
        http_request = _to_http_request(self.url, request)

        @fail_to(answer)
        def on_fetch(future):
            if future.exception():
                _handle_request_exception(future.exception())

            result = future.result()
            answer.set_result(
                Response(
                    headers=headers.from_http_headers(result.headers),
                    body=result.body,
                ),
            )

        self.client.fetch(http_request).add_done_callback(on_fetch)
        return answer


def _to_http_request(url, request):
    http_headers = headers.to_http_headers(request.headers or {})
    http_headers.update({
        headers.CALLER: request.caller or '',
        headers.SERVICE: request.service or '',
        headers.ENCODING: request.encoding or '',
        headers.TTL: '10000',  # TODO
        headers.PROCEDURE: request.procedure or '',
    })

    http_request = HTTPRequest(
        url=url,
        method='POST',
        headers=http_headers,
        body=request.body or '',
        # request_timeout=None,  TODO
    )
    return http_request


def _handle_request_exception(e):
    if not isinstance(e, HTTPError):
        reraise(e.exc_info())
    message = e.response.body
    if message.endswith('\n'):
        message = message[:-1]
    if 400 <= e.code < 500:
        raise BadRequestError(message)
    raise UnexpectedError(message)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)


from yarpc.errors import NoOutboundForServiceError, UnrecognizedProcedureError
from yarpc.transport import Channel


class RPC(object):

    def __init__(self, service, inbounds=None, outbounds=None):
        self.service = service
        self._inbounds = inbounds
        self._outbounds = outbounds or {}
        self._dispatcher = _Dispatcher()

    def register(self, registrant):
        # TODO decorator and register(procedure, handler) versions.
        self._dispatcher.register(self.service, registrant)

    def start(self):
        for inbound in self._inbounds:
            inbound.start(self)

    def stop(self):
        for inbound in self._inbounds:
            inbound.stop()

    def handle(self, request):
        return self._dispatcher.handle(request)

    def channel(self, service):
        assert service, "service is required"

        outbound = self._outbounds.get(service)
        if outbound is None:
            raise NoOutboundForServiceError(
                "no configured outbound transport for service %s" % service
            )

        return Channel(
            caller=self.service,
            service=service,
            outbound=outbound,
        )


class _Dispatcher(object):
    """A Handler which allows registering multiple procedures."""

    def __init__(self):
        self._procedures = {}

    def register(self, service, registrant):
        for procedure, handler in registrant.procedures.items():
            self._procedures[(service, procedure)] = handler

    def handle(self, request):
        procedure = self._procedures.get((request.service, request.procedure))

        if procedure is None:
            raise UnrecognizedProcedureError(
                service=request.service,
                procedure=request.procedure,
            )

        # TODO enable returning just body, instead of request
        response = procedure.handle(request)
        return response



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module provides encoding and concurrency agnostic abstractions that
define how requests and responses can be encoded or decoded.

It defines an OutboundEncoder which can be used to encode outgoing requests
and decode their responses, and an InboundEncoder which can be used to decode
incoming requests and encode their responses.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from yarpc.errors import ServerEncodingError
from yarpc._future import reraise


class OutboundEncoder(object):
    """Encodes outgoing requests and decodes their responses.

    .. code-block:: python

        request = encoder.encode_request(request)
        response = send(request)
        response = encoder.decode_response(response)

    Encoders are scoped to single requests and should not be re-used across
    multiple requests. The same encoder that was used to encode a request MUST
    be used to decode its response.
    """

    __slots__ = ()

    def encode_request(self, request):
        """Encodes the given request with this encoding.

        encode_request MUST be called before decode_response.

        This MAY mutate the underlying request or return a completely new
        request object. Callers should assume that they have lost ownership of
        the request object.

        :param yarpc.Request request:
            Request to encode
        :returns yarpc.Request:
            Encoded request
        """
        raise NotImplementedError

    def decode_response(self, response):
        """Decodes the given response with this encoding.

        encode_request MUST be called before decode_response.

        This MAY mutate the underlying response or return a completely new
        response object. Callers should assume that they have lost ownership
        of the response object.

        :param yarpc.Response response:
            Response to encode
        :returns yarpc.Response:
            Encoded response
        """
        raise NotImplementedError


class InboundEncoder(object):
    """Decodes incoming requests and encodes their responses or failures in
    handling the requests.

    .. code-block:: python

        try:
            request = encoder.decode_request(request)
            response = handle(request)
        except Exception:
            response = encoder.encode_exc_info(sys.exc_info())
        else:
            response = encoder.encode_response(response)

    Encoders are scoped to single requests and should not be re-used across
    multiple requests. The same encoder that was used to decode a request MUST
    be used to encode its response.

    Encoding errors MUST raise ``yarpc.errors.ServerEncodingError`` in
    order for the correct error messaging to be sent across the wire.
    """

    __slots__ = ()

    def decode_request(self, request):
        """Decodes the given request with this encoding.

        decode_request MUST be called before encode_response.

        This MAY mutate the underlying request or return a completely new
        request object. Callers should assume that they have lost ownership of
        the request object.

        :param yarpc.Request request:
            Request to encode
        :returns yarpc.Request:
            Encoded request
        """
        raise NotImplementedError

    def encode_response(self, response):
        """Encodes the given response with this encoding.

        decode_request MUST be called before encode_response. encode_response
        and encode_exc_info SHOULD return the same kind of object.

        This MAY mutate the underlying response or return a completely new
        response object. Callers should assume that they have lost ownership
        of the response object.

        :param yarpc.Response response:
            Response to encode
        :returns yarpc.Response:
            Encoded response
        """
        raise NotImplementedError

    def encode_exc_info(self, exc_info):
        """Encodes the given failure into a response.

        decode_request MUST be called before encode_exc_info. encode_response
        and encode_exc_info SHOULD return the same kind of object.

        This function MAY handle the exception and return a response. If it is
        unable to handle the exception, it MUST re-raise the exception.

        The default behavior is to re-raise the exception.

        :param tuple exc_info:
            Triple of ``(class, exception, traceback)`` representing the
            failure.
        :returns yarpc.Response:
            Encoded response
        """
        reraise(exc_info)


class SimpleEncoder(OutboundEncoder, InboundEncoder):
    """SimpleEncoder is an encoding for the simple case where request and
    response bodies are encoded in the same way and no local state is
    required.

    :param name:
        Name of the encoding.
    :param encode:
        A function that accepts the request/response body and returns bytes.
    :param decode:
        A function that accepts bytes and returns the request/response body.
    """

    __slots__ = ('name', 'encode', 'decode')

    def __init__(self, name, encode, decode):
        self.name = name
        self.encode = encode
        self.decode = decode

    def __str__(self):
        return 'SimpleEncoder(%s)' % self.name

    __repr__ = __str__

    def encode_request(self, request):
        request.encoding = self.name
        request.body = self.encode(request.body)
        return request

    def decode_request(self, request):
        request.body = self.decode(request.body)
        return request

    def encode_response(self, request, response):
        try:
            response.body = self.encode(response.body)
        except Exception as e:
            raise ServerEncodingError(
                exception=e,
                encoding=self.name,
                service=request.service,
                procedure=request.procedure,
                caller=request.caller,
            )
        return response

    def decode_response(self, response):
        response.body = self.decode(response.body)
        return response



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import socket
import time

from tchannel import TChannel
from tornado import gen

from yarpc import Response, Request
from yarpc.encoding import json
from yarpc.transport import Channel
from yarpc.transport.http import HTTPOutbound
from yarpc.transport.tchannel import TChannelOutbound

hostname = socket.gethostname()


@json.procedure('phone')
@gen.coroutine
def phone(request):
    start = time.time()

    # enforce request shape
    if (
        not request.body.get('transport') or
        not request.body.get('service') or
        not request.body.get('procedure')
    ):
        raise ValueError(
            'transport, service, and procedure are '
            'required request params'
        )

    # create outbound dynamically based on request
    t = request.body['transport']
    if 'http' in t:
        http = t['http']
        outbound = HTTPOutbound(
            url='http://%s:%d' % (http['host'], http['port']),
        )
    elif 'tchannel' in request.body['transport']:
        tch = t['tchannel']
        outbound = TChannelOutbound(
            tchannel=TChannel('yarpc-test'),
            hostport='%s:%d' % (tch['host'], tch['port']),
        )
    else:
        raise ValueError(
            "no transport available for: %s" %
            request.body['transport']
        )

    # create a client with a custom channel
    # based on the previously created outbound
    # TODO support arbitrary bytes and encodings
    client = json.JSONClient(Channel(
        caller='yarpc-test',
        service=request.body['service'],
        outbound=outbound,
    ))
    answer = yield client.call(Request(
        procedure=request.body['procedure'],
        body=request.body,
        ttl=10000,
    ))

    # respond with answer from downstream service
    response = Response(body={
        'service': request.body['service'],
        'procedure': request.body['procedure'],
        'hostname': hostname,
        'elapsedms': to_millis(time.time() - start),
        'body': answer.body,
    })

    raise gen.Return(response)


def to_millis(time):
    return int(round(time * 1000))
